import numpy as np
from dask.dataframe.extensions import make_array_nonempty
from pandas.core.dtypes.dtypes import register_extension_dtype

from ..geometry.line import Line, LineArray, LineDtype


@register_extension_dtype
class RingDtype(LineDtype):
    _geometry_name = 'ring'

    @classmethod
    def construct_array_type(cls, *args):
        return RingArray


class Ring(Line):

    @classmethod
    def construct_array_type(cls):
        return RingArray

    def to_shapely(self):
        """
        Convert to shapely shape

        Returns:
            shapely LinearRing shape
        """
        import shapely.geometry as sg
        line_coords = np.array(self.data.as_py(), dtype=self.numpy_dtype)
        return sg.LinearRing(line_coords.reshape(len(line_coords) // 2, 2))

    @classmethod
    def from_shapely(cls, shape):
        """
        Build a spatialpandas Ring object from a shapely shape

        Args:
            shape: A shapely LinearRing shape

        Returns:
            spatialpandas Ring
        """
        return super().from_shapely(shape)


class RingArray(LineArray):
    _element_type = Ring

    @property
    def _dtype_class(self):
        return RingDtype

    @classmethod
    def from_geopandas(cls, ga):
        """
        Build a spatialpandas RingArray from a geopandas GeometryArray or
        GeoSeries.

        Args:
            ga: A geopandas GeometryArray or GeoSeries of LinearRing shapes.

        Returns:
            RingArray
        """
        return super().from_geopandas(ga)


def _ring_array_non_empty(dtype):
    """
    Create an example length 2 array to register with Dask.
    See https://docs.dask.org/en/latest/dataframe-extend.html#extension-arrays
    """
    return RingArray([
        [0, 0, 1, 0, 1, 1, 0, 0],
        [2, 2, 2, 3, 3, 3, 2, 2]
    ], dtype=dtype)


if make_array_nonempty:
    make_array_nonempty.register(RingDtype)(_ring_array_non_empty)



from math import sqrt

import numpy as np

from ...utils import ngjit


@ngjit
def compute_line_length(values, value_offsets):
    total_len = 0.0
    for offset_ind in range(len(value_offsets) - 1):
        start = value_offsets[offset_ind]
        stop = value_offsets[offset_ind + 1]
        x0 = values[start]
        y0 = values[start + 1]

        for i in range(start + 2, stop, 2):
            x1 = values[i]
            y1 = values[i + 1]

            if (np.isfinite(x0) and np.isfinite(y0) and
                    np.isfinite(x1) and np.isfinite(y1)):
                total_len += sqrt((x1 - x0) ** 2 + (y1 - y0) ** 2)

            x0 = x1
            y0 = y1

    return total_len


@ngjit
def compute_area(values, value_offsets):
    area = 0.0

    for offset_ind in range(len(value_offsets) - 1):
        start = value_offsets[offset_ind]
        stop = value_offsets[offset_ind + 1]
        poly_length = stop - start

        if poly_length < 6:
            # A degenerate polygon, zero area
            continue

        for k in range(start, stop - 4, 2):
            i, j = k + 2, k + 4
            ix = values[i]
            jy = values[j + 1]
            ky = values[k + 1]

            area += ix * (jy - ky)

        # wrap-around term for polygon
        firstx = values[start]
        secondy = values[start + 3]
        lasty = values[stop - 3]
        area += firstx * (secondy - lasty)

    return area / 2.0



import numpy as np

from spatialpandas.geometry import (
    Line,
    LineArray,
    MultiPoint,
    MultiPointArray,
    MultiPolygon,
    MultiPolygonArray,
    Point,
    PointArray,
    Polygon,
    PolygonArray,
    Ring,
)

unit_square_cw = np.array([1, 1,  1, 2,  2, 2,  2, 1,  1, 1], dtype='float64')
large_square_ccw = np.array([0, 0, 3, 0, 3, 3, 0, 3, 0, 0], dtype='float64')


def test_point():
    point = Point([1, 2])
    assert point.length == 0.0
    assert point.area == 0.0


def test_point_array():
    points = PointArray(unit_square_cw)
    np.testing.assert_equal(points.length, [0.0, 0.0, 0.0, 0.0, 0.0])
    np.testing.assert_equal(points.area, [0.0, 0.0, 0.0, 0.0, 0.0])
    assert points.total_bounds == (1, 1, 2, 2)


def test_multipoint():
    points = MultiPoint(unit_square_cw)
    assert points.length == 0.0
    assert points.area == 0.0


def test_multipoint_array():
    points = MultiPointArray([
        unit_square_cw,
        large_square_ccw,
        np.concatenate([large_square_ccw, unit_square_cw])
    ])

    np.testing.assert_equal(points.length, [0.0, 0.0, 0.0])
    np.testing.assert_equal(points.area, [0.0, 0.0, 0.0])
    assert points.total_bounds == (0.0, 0.0, 3.0, 3.0)


def test_lines():
    lines = Line(unit_square_cw)
    assert lines.length == 4.0
    assert lines.area == 0.0


def test_lines_array():
    lines = LineArray([
        unit_square_cw,
        large_square_ccw,
        np.concatenate([large_square_ccw, [np.nan, np.nan], unit_square_cw])
    ])

    np.testing.assert_equal(lines.length, [4.0, 12.0, 16.0])
    np.testing.assert_equal(lines.area, [0.0, 0.0, 0.0])
    assert lines.total_bounds == (0.0, 0.0, 3.0, 3.0)


def test_ring():
    lines = Ring(unit_square_cw)
    assert lines.length == 4.0
    assert lines.area == 0.0


def test_ring_array():
    lines = LineArray([
        unit_square_cw,
        large_square_ccw,
    ])

    np.testing.assert_equal(lines.length, [4.0, 12.0])
    np.testing.assert_equal(lines.area, [0.0, 0.0])
    assert lines.total_bounds == (0.0, 0.0, 3.0, 3.0)


def test_polygon():
    polygon = Polygon([large_square_ccw, unit_square_cw])
    assert polygon.length == 16.0
    assert polygon.area == 8.0


def test_polygon_array():
    polygons = PolygonArray([
        [large_square_ccw],
        [large_square_ccw, unit_square_cw],
        [unit_square_cw]
    ])
    np.testing.assert_equal(polygons.length, [12.0, 16.0, 4.0])
    np.testing.assert_equal(polygons.area, [9.0, 8.0, -1.0])
    assert polygons.total_bounds == (0.0, 0.0, 3.0, 3.0)


def test_multipolygon():
    multipolygon = MultiPolygon([
        [large_square_ccw, unit_square_cw],
        [large_square_ccw + 4.0]
    ])
    assert multipolygon.length == 28.0
    assert multipolygon.area == 17.0


def test_multipolygon_array():
    multipolygon = MultiPolygonArray([
        [
            [large_square_ccw, unit_square_cw],
            [large_square_ccw + 4.0]
        ], [
            [large_square_ccw + 8.0]
        ]
    ])
    np.testing.assert_equal(multipolygon.length, [28.0, 12.0])
    np.testing.assert_equal(multipolygon.area, [17.0, 9.0])
    assert multipolygon.total_bounds == (0.0, 0.0, 11.0, 11.0)



import numpy as np
from dask.dataframe.extensions import make_array_nonempty
from pandas.core.dtypes.dtypes import register_extension_dtype

from ..geometry._algorithms.intersection import multipoints_intersect_bounds
from ..geometry.base import GeometryDtype
from ..geometry.baselist import GeometryList, GeometryListArray


@register_extension_dtype
class MultiPointDtype(GeometryDtype):
    _geometry_name = 'multipoint'
    @classmethod
    def construct_array_type(cls, *args):
        if len(args) > 0:
            raise NotImplementedError("construct_array_type does not support arguments")
        return MultiPointArray


class MultiPoint(GeometryList):
    _nesting_levels = 0

    @classmethod
    def construct_array_type(cls):
        return MultiPointArray

    @classmethod
    def _shapely_to_coordinates(cls, shape):
        import shapely.geometry as sg
        if isinstance(shape, sg.Point):
            return np.asarray(shape.coords).ravel()
        elif isinstance(shape, sg.MultiPoint):
            return np.hstack([g.coords for g in shape.geoms]).ravel()
        else:
            raise ValueError("""
Received invalid value of type {typ}. Must be an instance of Point,
or MultiPoint""".format(typ=type(shape).__name__))

    def to_shapely(self):
        """
        Convert to shapely shape

        Returns:
            shapely MultiPoint shape
        """
        import shapely.geometry as sg
        point_coords = np.array(self.data.as_py(), dtype=self.numpy_dtype)
        return sg.MultiPoint(point_coords.reshape(len(point_coords) // 2, 2))

    @classmethod
    def from_shapely(cls, shape):
        """
        Build a spatialpandas MultiPoint object from a shapely shape

        Args:
            shape: A shapely MultiPoint or Point shape

        Returns:
            spatialpandas MultiPoint
        """
        return super().from_shapely(shape)

    @property
    def length(self):
        return 0.0

    @property
    def area(self):
        return 0.0

    def intersects_bounds(self, bounds):
        x0, y0, x1, y1 = bounds
        result = np.zeros(1, dtype=np.bool_)
        offsets = self.buffer_outer_offsets
        multipoints_intersect_bounds(
            float(x0), float(y0), float(x1), float(y1),
            self.buffer_values, offsets[:-1], offsets[1:], result
        )
        return result[0]


class MultiPointArray(GeometryListArray):
    _element_type = MultiPoint
    _nesting_levels = 1

    @property
    def _dtype_class(self):
        return MultiPointDtype

    @classmethod
    def from_geopandas(cls, ga):
        """
        Build a spatialpandas MultiPointArray from a geopandas GeometryArray or
        GeoSeries.

        Args:
            ga: A geopandas GeometryArray or GeoSeries of MultiPoint or
            Point shapes.

        Returns:
            MultiPointArray
        """
        return super().from_geopandas(ga)

    @property
    def length(self):
        return np.zeros(len(self), dtype=np.float64)

    @property
    def area(self):
        return np.zeros(len(self), dtype=np.float64)

    def intersects_bounds(self, bounds, inds=None):
        x0, y0, x1, y1 = bounds
        offsets0 = self.buffer_outer_offsets
        start_offsets0 = offsets0[:-1]
        stop_offsets0 = offsets0[1:]

        if inds is not None:
            start_offsets0 = start_offsets0[inds]
            stop_offsets0 = stop_offsets0[inds]

        result = np.zeros(len(start_offsets0), dtype=np.bool_)
        multipoints_intersect_bounds(
            float(x0), float(y0), float(x1), float(y1),
            self.buffer_values, start_offsets0, stop_offsets0, result
        )
        return result


def _multi_points_array_non_empty(dtype):
    """
    Create an example length 2 array to register with Dask.
    See https://docs.dask.org/en/latest/dataframe-extend.html#extension-arrays
    """
    return MultiPointArray([
        [1, 0, 1, 1],
        [1, 2, 0, 0]
    ], dtype=dtype)


if make_array_nonempty:
    make_array_nonempty.register(MultiPointDtype)(_multi_points_array_non_empty)



# Script to create parquet files that are read in test_parquet.
# The files can be created with particular version of pyarrow and dask, and
# checked that they can be read with different versions of them.

import dask.dataframe as dd
from hypothesis.errors import NonInteractiveExampleWarning
import os
from pyarrow import __version__ as pyarrow_version
from spatialpandas import GeoDataFrame
from spatialpandas.io import to_parquet
from spatialpandas.tests.geometry.strategies import st_multiline_array
import warnings


directory = "test_data"


def _generate_test_geodata():
    n = 5
    warnings.filterwarnings("ignore", category=NonInteractiveExampleWarning)
    lines = st_multiline_array(min_size=n, max_size=n, geoseries=True).example()
    return lines


def _write(use_dask=False):
    if use_dask:
        filename = f"dask_{pyarrow_version}.parq"
    else:
        filename = f"serial_{pyarrow_version}.parq"

    lines = _generate_test_geodata()
    df = GeoDataFrame({'multiline': lines, 'a': list(range(len(lines)))})

    path = os.path.join(directory, filename)

    if use_dask:
        ddf = dd.from_pandas(df, npartitions=2)
        ddf.to_parquet(path)
    else:
        to_parquet(df, path)


def _write_repartitioned():
    filename = f"dask_repart_{pyarrow_version}.parq"

    lines = _generate_test_geodata()
    df = GeoDataFrame({'multiline': lines, 'a': list(range(len(lines)))})
    ddf = dd.from_pandas(df, npartitions=2)

    path = os.path.join(directory, filename)

    ddf.pack_partitions_to_parquet(path, npartitions=2, overwrite=True)


if __name__ == '__main__':
    _write(False)
    _write(True)
    _write_repartitioned()



import numpy as np

from ...geometry._algorithms.measures import compute_area
from ...utils import ngjit


@ngjit
def triangle_orientation(ax, ay, bx, by, cx, cy):
    """
    Calculate orientation of triangle
    Args:
        ax, ay: coords of first point
        bx, by: coords of second point
        cx, cy: coords of third point

    Returns:
        +1 if counter clockwise
         0 if colinear
        -1 if clockwise
    """
    ab_x, ab_y = bx - ax, by - ay
    ac_x, ac_y = cx - ax, cy - ay

    # compute cross product: ab x bc
    ab_x_ac = (ab_x * ac_y) - (ab_y * ac_x)

    if ab_x_ac > 0:
        # Counter clockwise
        return 1
    elif ab_x_ac < 0:
        # Clockwise
        return -1
    else:
        # Collinear
        return 0


@ngjit
def orient_polygons(values, polygon_offsets, ring_offsets):
    """
    Orient polygons so that exterior is in CCW order and interior rings (holes) are in
    CW order.

    This function mutates the values array

    Args:
        values: Ring coordinates
        polygon_offsets: Offsets into ring_offsets of first ring in each polygon
        ring_offsets: Offsets into values of the start of each ring
    """
    num_rings = len(ring_offsets) - 1

    # Compute expected orientation of rings
    expected_ccw = np.zeros(len(ring_offsets) - 1, dtype=np.bool_)
    expected_ccw[polygon_offsets[:-1]] = True

    # Compute actual orientation of rings
    is_ccw = np.zeros(num_rings)
    for i in range(num_rings):
        is_ccw[i] = compute_area(values, ring_offsets[i:i + 2]) >= 0

    # Compute indices of rings to flip
    flip_inds = np.nonzero(is_ccw != expected_ccw)
    ring_starts = ring_offsets[:-1]
    ring_stops = ring_offsets[1:]
    flip_starts = ring_starts[flip_inds]
    flip_stops = ring_stops[flip_inds]

    for i in range(len(flip_starts)):
        flip_start = flip_starts[i]
        flip_stop = flip_stops[i]

        xs = values[flip_start:flip_stop:2]
        ys = values[flip_start + 1:flip_stop:2]
        values[flip_start:flip_stop:2] = xs[::-1]
        values[flip_start + 1:flip_stop:2] = ys[::-1]
